import os
import json
import datetime
import warnings
from collections import defaultdict, OrderedDict

'''length scale conversion to meter'''
LENGTH_SCALE_CONVERSION = {
    'millimeter': 1e-3,
    'meter': 1.0,
}

#############################################

def parse_point(json_point, scale=1.0):
    return [scale*json_point[0], scale*json_point[1], scale*json_point[2]]

def parse_nodes(node_data, scale=1.0):
    nodes = []
    for n in node_data:
        n['point'] = parse_point(n['point'], scale=scale)
        nodes.append(Node.from_data(n))
    return nodes

def extract_model_name_from_path(file_path):
    forename = file_path.split('.json')[0]
    return forename.split(os.sep)[-1]

#############################################

class Model(object):
    def __init__(self, nodes, elements, supports, joints, materials, crosssecs, unit='meter', model_name=None):
        self.model_name = model_name
        self.generate_time = str(datetime.datetime.now())
        assert unit in LENGTH_SCALE_CONVERSION
        self.unit = unit
        scale = LENGTH_SCALE_CONVERSION[unit]
        if scale != 1.0:
            warnings.warn('Model unit {}, scaled by {} and converted to Meter unit.'.format(unit, scale))
        for node in nodes:
            node.point = [scale*cp for cp in node.point]

        self.nodes = nodes
        self.elements = elements

        # turn lists into dicts
        self.supports = {}
        for support in supports:
            self.supports[support.node_ind] = support

        self.joints = {}
        if joints is not None:
            for joint in joints:
                for e_tag in joint.elem_tags:
                    if e_tag in self.joints:
                        warnings.warn('Multiple joints assigned to the same element tag |{}|!'.format(e_tag))
                    self.joints[e_tag] = joint

        self.materials = {}
        for mat in materials:
            for e_tag in mat.elem_tags:
                if e_tag in self.materials:
                    warnings.warn('Multiple materials assigned to the same element tag |{}|!'.format(e_tag))
                self.materials[e_tag] = mat

        self.crosssecs = {}
        for cs in crosssecs:
            for e_tag in cs.elem_tags:
                if e_tag in self.crosssecs:
                    warnings.warn('Multiple materials assigned to the same element tag |{}|!'.format(e_tag))
                self.crosssecs[e_tag] = cs

    @property
    def node_num(self):
        return len(self.nodes)

    @property
    def element_num(self):
        return len(self.elements)

    @classmethod
    def from_json(cls, file_path, verbose=False):
        assert os.path.exists(file_path) and "json file path does not exist!"
        with open(file_path, 'r') as f:
            json_data = json.loads(f.read())
        if 'model_name' not in json_data:
            json_data['model_name']  = extract_model_name_from_path(file_path)
        return cls.from_data(json_data, verbose)

    @classmethod
    def from_data(cls, data, verbose=False):
        model_name = data['model_name']
        unit = data['unit']
        # length scale for vertex positions
        scale = LENGTH_SCALE_CONVERSION[unit]
        unit = 'meter'
        nodes = parse_nodes(data['nodes'], scale=scale)
        elements = [Element.from_data(e) for e in data['elements']]
        element_inds_from_tag = defaultdict(list)
        for e in elements:
            element_inds_from_tag[e.elem_tag].append(e.elem_ind)
        supports = [Support.from_data(s) for s in data['supports']]
        joints = [Joint.from_data(j) for j in data['joints']]
        materials = [Material.from_data(m) for m in data['materials']]
        crosssecs = [CrossSec.from_data(c) for c in data['cross_secs']]
        # sanity checks
        if 'node_num' in data:
            assert len(nodes) == data['node_num']
        if 'element_num' in data:
            assert len(elements) == data['element_num']
        node_id_range = list(range(len(nodes)))
        for e in elements:
            assert e.end_node_inds[0] in node_id_range and \
                   e.end_node_inds[1] in node_id_range, \
                   'element end point id not in node_list id range!'
        num_grounded_nodes = 0
        for n in nodes:
            if n.is_grounded:
                num_grounded_nodes += 1
        # assert grounded_nodes > 0, 'The structure must have at lease one grounded node!'
        if num_grounded_nodes == 0:
            warnings.warn('The structure must have at lease one grounded node!')
        if verbose:
            print('Model: {} | Original unit: {} | Generated time: {}'.format(model_name, data['unit'], 
                data['generate_time'] if 'generate_time' in data else ''))
            print('Nodes: {} | Elements: {} | Supports: {} | Joints: {} | Materials: {} | Cross Secs: {} | Tag Ground: {} '.format(
                len(nodes), len(elements), len(supports), len(joints), len(materials), len(crosssecs), num_grounded_nodes))
        return cls(nodes, elements, supports, joints, materials, crosssecs, model_name=model_name)

    def to_data(self):
        data = OrderedDict()
        data['model_name'] = self.model_name
        data['unit'] = self.unit
        data['generate_time'] = self.generate_time
        data['node_num'] = len(self.nodes)
        data['element_num'] = len(self.elements)
        data['nodes'] = [n.to_data() for n in self.nodes]
        data['elements'] = [e.to_data() for e in self.elements]
        data['supports'] = [s.to_data() for s in self.supports.values()]
        data['joints'] = [j.to_data() for j in self.joints.values()]
        data['materials'] = [m.to_data() for m in self.materials.values()]
        data['cross_secs'] = [cs.to_data() for cs in self.crosssecs.values()]
        return data

class LoadCase(object):
    def __init__(self, point_loads=None, uniform_element_loads=None, gravity_load=None):
        self.point_loads = point_loads or []
        self.uniform_element_loads = uniform_element_loads or []
        self.gravity_load = gravity_load

    @classmethod
    def from_json(cls, file_path, lc_ind=0):
        assert os.path.exists(file_path) and "json file path does not exist!"
        with open(file_path, 'r') as f:
            json_data = json.loads(f.read())
        json_data = json_data['loadcases'] if 'loadcases' in json_data else json_data
        return cls.from_data(json_data[str(lc_ind)])
    
    @classmethod
    def from_data(cls, data):
        point_loads = [PointLoad.from_data(pl) for pl in data['ploads']]
        uniform_element_loads = [UniformlyDistLoad.from_data(el) for el in data['eloads']]
        gravity_load = None if 'gravity' not in data else data['gravity']
        return cls(point_loads, uniform_element_loads, gravity_load)

    def to_data(self):
        data = {'ploads' : [pl.to_data() for pl in self.point_loads],
                'eloads' : [el.to_data() for el in self.uniform_element_loads],
                'gravity' : self.gravity_load.to_data() if self.gravity_load is not None else None
                }
        return data

    def __repr__(self):
        return '{}(#pl:{},#el:{},gravity:{})'.format(self.__class__.__name__, len(self.point_loads), 
            len(self.uniform_element_loads), self.gravity_load)

##############################################

class Node(object):
    def __init__(self, point, node_ind, is_grounded):
        self.point = point
        self.node_ind = node_ind
        self.is_grounded = is_grounded

    @classmethod
    def from_data(cls, data):
        return cls(data['point'], data['node_ind'], data['is_grounded'])

    def to_data(self):
        data = {'point' : list(self.point), 'node_ind' : self.node_ind, 'is_grounded' : self.is_grounded}
        return data

    def __repr__(self):
        return '{}(#{},{},Grd:{})'.format(self.__class__.__name__, self.node_ind, self.point, self.is_grounded)

class Element(object):
    def __init__(self, end_node_inds, elem_ind, elem_tag='', bending_stiff=True):
        assert end_node_inds[0] != end_node_inds[1], 'zero length element not allowed!'
        self.end_node_inds = end_node_inds
        self.elem_tag = elem_tag
        self.elem_ind = elem_ind
        self.bending_stiff = bending_stiff

    @classmethod
    def from_data(cls, data):
        return cls(data['end_node_inds'], data['elem_ind'], data['elem_tag'], data['bending_stiff'])

    def to_data(self):
        data = {'end_node_inds' : self.end_node_inds,
                'elem_ind' : self.elem_ind,
                'elem_tag' : self.elem_tag,
                'bending_stiff' : self.bending_stiff,
                }
        return data

    def __repr__(self):
        return '{}(#{}({}),{},Bend:{})'.format(self.__class__.__name__, self.elem_ind, self.elem_tag, self.end_node_inds, self.bending_stiff)

class Support(object):
    def __init__(self, condition, node_ind):
        self.condition = condition
        self.node_ind = node_ind

    @classmethod
    def from_data(cls, data):
        return cls(data['condition'], data['node_ind'])

    def to_data(self):
        return {
            'condition' : self.condition,
            'node_ind' : self.node_ind,
        }

    def __repr__(self):
        return '{}(#{},{})'.format(self.__class__.__name__, self.node_ind, self.condition)

class Joint(object):
    def __init__(self, c_conditions, elem_tags):
        assert len(c_conditions) == 12
        self.c_conditions = c_conditions
        self.elem_tags = elem_tags

    @classmethod
    def from_data(cls, data):
        return cls(data['c_conditions'], data['elem_tags'])

    def to_data(self):
        return {
            'c_conditions' : self.c_conditions,
            'elem_tags' : self.elem_tags,
        }

# TODO: assumed to be converted to meter-based unit
class CrossSec(object):
    def __init__(self, A, Jx, Iy, Iz, elem_tags=None, family='unnamed', name='unnamed'):
        self.A = A
        self.Jx = Jx
        self.Iy = Iy
        self.Iz = Iz
        self.elem_tags = elem_tags if elem_tags else [None]
        self.family = family
        self.name = name

    @classmethod
    def from_data(cls, data):
        return cls(data['A'], data['Jx'], data['Iy'], data['Iz'], data['elem_tags'], data['family'], data['name'])

    def to_data(self):
        return {
            'A' : self.A,
            'Jx' : self.Jx,
            'Iy' : self.Iy,
            'Iz' : self.Iz,
            'elem_tags' : self.elem_tags,
            'family' : self.family,
            'name' : self.name,
        }

    def __repr__(self):
        return '{}(family:{} name:{} area:{}[m2] Jx:{}[m4] Iy:{}[m4] Iz:{}[m4] applies to elements:{})'.format(
            self.__class__.__name__, self.family, self.name, self.A, self.Jx, self.Iy, self.Iz, self.elem_tags)

def mu2G(mu, E):
    """compute shear modulus from poisson ratio mu and Youngs modulus E
    """
    return E/(2*(1+mu))

def G2mu(G, E):
    """compute poisson ratio from shear modulus and Youngs modulus E
    """
    return E/(2*G)-1

# TODO: assumed to be converted to kN/m2, kN/m3
class Material(object):
    def __init__(self, E, G12, fy, density, elem_tags=None, family='unnamed', name='unnamed', type_name='ISO', G3=None):
        self.E = E
        # in-plane shear modulus 
        self.G12 = G12
        # transverse shear modulus
        self.G3 = G3 or G12
        self.mu = G2mu(G12, E)
        # material strength in the specified direction (local x direction)
        self.fy = fy
        self.density = density
        self.elem_tags = elem_tags if elem_tags else [None]
        self.family = family
        self.name = name
        self.type_name = type_name

    @classmethod
    def from_data(cls, data):
        return cls(data['E'], data['G12'], data['fy'], data['density'], data['elem_tags'], data['family'], data['name'], data['type_name'], data.get('G3', None))

    def to_data(self):
        return  {
            'E' : self.E,
            'G12' : self.G12,
            'G3' : self.G3,
            'mu' : self.mu,
            'fy' : self.fy,
            'density' : self.density,
            'elem_tags' : self.elem_tags,
            'family' : self.family,
            'name' : self.name,
            'type_name' : self.type_name,
        }

    def __repr__(self):
        # G3:8076[kN/cm2]
        return '{}(|{}| E:{}[kN/m2] mu:{} G12:{}[kN/m2] G3:{}[kN/m2] density:{}[kN/m3] fy:{}[kN/m2] applies to elements:{})'.format(
            self.__class__.__name__, self.family+'-'+self.name, self.E, self.mu, self.G12, self.G3, self.density, self.fy, self.elem_tags)

class PointLoad(object):
    def __init__(self, force, moment, node_ind, loadcase=0):
        self.force = force
        self.moment = moment
        self.node_ind = node_ind
        self.loadcase = loadcase

    @classmethod
    def from_data(cls, data):
        lc_ind = 0 if 'loadcase' not in data else data['loadcase']
        return cls(data['force'], data['moment'], data['node_ind'], lc_ind)

    def to_data(self):
        return {
            'force' : self.force,
            'moment' : self.moment,
            'node_ind' : self.node_ind,
            'loadcase' : self.loadcase,
        }

    def __repr__(self):
        return '{}(node_ind {} | force {} | moment {} | lc#{})'.format(self.__class__.__name__, self.node_ind, self.force, self.moment, self.loadcase)

class UniformlyDistLoad(object):
    def __init__(self, q, load, elem_tags, loadcase=0):
        self.q = q
        # not sure if load is used in Karamba, we only use q here
        self.load = load
        self.elem_tags = elem_tags
        self.loadcase = loadcase

    @classmethod
    def from_data(cls, data):
        lc_ind = 0 if 'loadcase' not in data else data['loadcase']
        return cls(data['q'], data['load'], data['elem_tags'], lc_ind)

    def to_data(self):
        return {
            'q' : self.q,
            'load' : self.load,
            'elem_tags' : self.elem_tags,
            'loadcase' : self.loadcase,
        }

    def __repr__(self):
        return '{}(element_tags {} | q {} | load {} | lc#{})'.format(self.__class__.__name__, 
            self.elem_tags, self.q, self.load, self.loadcase)

class GravityLoad(object):
    def __init__(self, force=[0,0,-1], loadcase=0):
        self.force = force
        self.loadcase = loadcase

    @classmethod
    def from_data(cls, data):
        lc_ind = 0 if 'loadcase' not in data else data['loadcase']
        return cls(data['force'], lc_ind)

    def to_data(self):
        return {
            'force' : self.force,
            'loadcase' : self.loadcase,
        }

    def __repr__(self):
        return '{}({}, lc#{})'.format(self.__class__.__name__, self.force, self.loadcase)

##########################################

class AnalysisResult(object):
    def __init__(self):
        pass


import numpy as np
import scipy
import scipy.sparse.linalg as SPLA
import numpy.linalg as LA
from numpy.linalg import norm, solve
from numpy.polynomial.polynomial import polyvander, polyval, Polynomial, polyfit

def interp_poly(d_u, d_v, L):
    """compute shape polynomial coeff for local x,y,z based on end nodal displacement value

    Parameters
    ----------
    d_u : 6x1 np array
        nodal u's displacement in local coordinate
    d_v : 6x1 np array
    L : float
        element length

    Return
    ------
    u_poly : (3x4) np array
        each row is polynomail coeff vector
    """
    D_local = np.hstack([d_u, d_v])
    dx_u, dx_v = (D_local[0], D_local[6])
    # * linear monomials 1+x
    A_l = polyvander([dx_u, L+dx_v], 1)

    # * cubic monomials 1, x, x**2, x**3
    A_c = np.zeros((4, 4))
    A_c[[0,2], :] = polyvander([dx_u, L+dx_v], 3)
    # c = np.hstack([Polynomial([1]*4).deriv().coef, [0]])
    # A[[1,3], :] = polyvander([0, L], 3) * c # apply entrywise mul on each row
    # * derivative of the monomials
    c = np.array([1,2,3])
    A_c[[1,3], 1:] = polyvander([dx_u, L+dx_v], 2) * c

    # polynomial evaluated at node pts
    # notice the negative sign here to make sign convention agree
    # coefficents of the polynomial functions under monomial basis
    d_x = [dx_u, dx_v]
    d_y = [D_local[1], D_local[5], D_local[7], D_local[11]]
    d_z = [D_local[2], -D_local[4], D_local[8], -D_local[10]]
    u_poly = [solve(A_l, d_x),
              solve(A_c, d_y),
              solve(A_c, d_z)]
    assert np.allclose(u_poly[0], polyfit([dx_u, L+dx_v], d_x, 1))
    return u_poly
    

def get_element_shape_fn(end_u, end_v, d_u, d_v, exagg=1.0):
    """cubic polynomial interpolation given its boundary condition:
        d = [u1, du1/dx, u2, du2/dx]
    
    Parameters
    ----------
    d : [type]
        [description]
    L : float
        element length

    Return
    ------
    poly_eval_fn : function handle
        shape_fn(t) -> np array of size (3,), node pts in global coordinate
        time parameter reparameterized to [0,1.0]
    """
    L = norm(end_u - end_v)
    R3 = global2local_transf_matrix(end_u, end_v)
    assert np.allclose((R3.T).dot(R3), np.eye(3))
    assert np.allclose((R3).dot(R3.T), np.eye(3))

    R = turn_diagblock(R3)
    # compute end deflection in local ordinates
    D_global = np.hstack([d_u, d_v])
    D_local = exagg * R.dot(D_global)

    dx_u, dx_v = (D_local[0], D_local[6])
    u_poly = interp_poly(D_local[:6], D_local[6:12], L)

    def shape_fn(t):
        assert t>=0 and t<=1.0
        d = np.array([polyval(dx_u + t*(L+dx_v-dx_u), u_poly[i]) for i in range(3)])
        pt_t = end_u + np.array([t*L, 0, 0])
        return (pt_t + R3.T.dot(d))
    return shape_fn


def get_internal_reaction_fn(fu, fv):
    # TODO cubic interpolation when in-span load is applied
    # * linear monomials 1+x
    A_l = polyvander([0., 1.], 1)
    # * cubic monomials 1, x, x**2, x**3
    A_c = np.zeros((4, 4))
    A_c[[0,2], :] = polyvander([0., 1.], 3)
    # * derivative of the monomials
    c = np.array([1,2,3])
    A_c[[1,3], 1:] = polyvander([0., 1.], 2) * c

    # polynomial evaluated at node pts
    # notice the negative sign here to make sign convention agree
    # coefficents of the polynomial functions under monomial basis
    # - linear interpolation for Nx, Fy, Fz, Tx
    u_poly = [solve(A_l, [fu[i], fv[i]]) for i in range(6)]
    # - cubic interpolation for moments, shear as tangents
    # Mz = [fu[5], -fu[1], fv[5], -fv[1]]
    # My = [fu[4], fu[2], fv[4], fv[2]]
    # u_poly.extend([solve(A_c, My), solve(A_c, Mz)])

    # Vz = dMy/dx 
    # assert np.allclose(-u_poly[1], Polynomial(u_poly[5]).deriv().coef)
    print((-fu[1], Polynomial(u_poly[5]).deriv().coef))
    # -Vy = dMz/dx 
    # assert np.allclose(u_poly[2], Polynomial(u_poly[4]).deriv().coef)
    print((-u_poly[2], Polynomial(u_poly[4]).deriv().coef))

    def local_force_fn(t):
        assert t>=0 and t<=1.0
        return np.array([polyval(t, u_poly[i]) for i in range(6)])
    return local_force_fn

###############################################


import numpy as np
from numpy.linalg import norm
from scipy.sparse import csc_matrix
from termcolor import cprint

from pyconmech.frame_analysis.io_base import mu2G, G2mu

DOUBLE_EPS = 1e-14
DOUBLE_INF = 1e14

def axial_stiffness_matrix(L, A, E, cr1=np.inf, cr2=np.inf):
    """local stiffness matrix for a bar with only axial force at two ends

            AE/L (u1 - u2) = F1
            AE/L (u2 - u1) = F2

        Virtual displacement (principle of virtual work) derivation:
        (same result can be derived using variational method or Galerkin method,
        which generalizes FEA to problems beyond structural mechanics.)

        We use linear (C0) interpolation between (x1, phi1) and ((x1, phi2))
        Thus the polynimial basis is [1, x]'
        The interpolation polynomial gives:
            [phi1, phi2]' = A [a0, a1]'
        where A = [[1, x1], [1, x2]]
        Inverting A gives:
            Ainv = (1/(x2-x1)) [[x2, -x1], [-1, 1]]
        Thus the *shape function* N:
            N = [(x2-x)/(x2-x1), (x-x1)/(x2-x1)]

        Thus, the displacement u can be interpolated over an element:
            u = N d
        where d is the nodal displacement dof of the element.

        Strains are determined from displacements:
            \epsilon = B d
        where [B] = [\partial][N] = d/dx [(L-x)/L, x/L]= [-1/L, L]

        Thus, the stiffness matrix 
            [k] = \int_0^L [B]^T E [B] A dx = AE [[1, -1], [-1, 1]]
    
    Parameters
    ----------
    L : [type]
        bar element length
    A : [type]
        cross section area
    E : [type]
        Young's modulus
    cr1: float
        rotational stiffness at the first end of the beam, default to infinite (rigid joint)
    cr2: float
        rotational stiffness at the second end of the beam, default to infinite (rigid joint)

    Return
    ------
    K_ax_x : 2x2 numpy array
    """
    K = np.ones([2,2])
    K[0,1] = -1.
    K[1,0] = -1.
    d_inv = 0.0
    if abs(cr1) < DOUBLE_EPS or abs(cr2) < DOUBLE_EPS:
        # cr1 = cr2 = 0, fully released
        d_inv = 0.0
    else:
        cr1_inv = 1/cr1 if cr1<DOUBLE_INF else 0.0
        cr2_inv = 1/cr2 if cr2<DOUBLE_INF else 0.0
        d_inv = 1.0 / (L/(E*A) + cr1_inv + cr2_inv) 
    K *= d_inv
    return K


def torsional_stiffness_matrix(L, J, G, cr1=np.inf, cr2=np.inf):
    """[summary]

    [Mx1, Mx2]' = ((GJ)/L) * [[1,-1], [-1,1]] [theta_x1, theta_x2]
    
    Parameters
    ----------
    L : [type]
        bar element length
    J : [type]
        torsional constant, unit: length unit^4
        In the case of circular, cylindrical shaft, it's equal to
        the polar moment of inertia of the cross section.
    G : [type]
        modulus of rigidity
    cr1: float
        rotational stiffness at the first end of the beam, default to infinite (rigid joint)
    cr2: float
        rotational stiffness at the second end of the beam, default to infinite (rigid joint)

    Return
    ------
    K_tor_x : 2x2 numpy array
    """
    return axial_stiffness_matrix(L, J, G, cr1=cr1, cr2=cr2)


def bending_stiffness_matrix(L, E, Iz, axis=2, cr1=np.inf, cr2=np.inf):
    """stiffness matrix of uniform beam element without the transverse shear deformation

    Here the stress-strain (\sigma ~ \epsilon) turns into 
    bending moment-curvature (M ~ \kappa).

        strain e_x = - y/\kappa = - y (d^2 v / d^2 x)
    where \kappa is the radius of curvature

    Then from \sigma_x = E e_x,
        \sigma_x = - E y (d^2 v / d^2 x)
    
    From moment equilibrium,
        M_z = - \int_A \sigma_x y dA = E Iz (d^2 v / d^2 x)

    where the lateral displacement v(x) = [N]{d} and \kappa = [B] {d}
        nodal dof {d} = [v_1, theta_z1, v2, theta_z2]
        [B] = d^2/d^2 x [N]

    We use cubic curve interpolation (C1) and [N]'s each row represents
    the Langrange's interpolation functions (polynomials of deg 3 in this case).
    Thus, the stiffness matrix [k] = int_0^L (E Iz d^2[N]) dx

    Parameters
    ----------
    L : [type]
        [description]
    E : [type]
        [description]
    Iz : [type]
        moment of inertia of the section about the z axis
            Iz = \int_A y^2 dA
    axis: int
        1 = local y axis , 2 = local z axis, default to 2
    cr1: float
        rotational stiffness at the first end of the beam, default to infinite (rigid joint)
    cr2: float
        rotational stiffness at the second end of the beam, default to infinite (rigid joint)

    Return
    ------
    K_bend_z : 4x4 numpy array
    """
    K = np.zeros([4,4])
    sign = 1. if axis==2 else -1.
    assert abs(E) > DOUBLE_EPS and abs(Iz) > DOUBLE_EPS
    a1 = cr1*L/E*Iz # k1 = a1 * EI/L
    a2 = cr2*L/E*Iz # k2 = a2 * EI/L
    # a = a1*a2 / (a1*a2 + 4*a1 + 4*a2 + 12)
    # aa = a * (1 + (a1+a2)/(a1*a2))
    # a2_2 = a * (1 + 2/a2)
    # a2_3 = a * (1 + 3/a2)
    # a1_2 = a * (1 + 2/a1)
    # a1_3 = a * (1 + 3/a1)
    if abs(a1) < DOUBLE_INF and abs(a2) < DOUBLE_INF:
        a_demo = a1*a2 + 4*a1 + 4*a2 + 12
        a = a1*a2 / a_demo
        aa = (a1*a2 + a1 +a2) / a_demo
        a2_2 = (a2 + 2)*a1 / a_demo
        a2_3 = (a2 + 3)*a1 / a_demo
        a1_2 = (a1 + 2)*a2 / a_demo
        a1_3 = (a1 + 3)*a2 / a_demo
    elif abs(a1) < DOUBLE_EPS and abs(a2) > DOUBLE_INF:
        # a1=0, a2 = inf
        # a = a1 / (a1 + 4)
        a = 0.0
        aa = 0.25
        a2_2 = 0.0
        a2_3 = 0.0
        a1_2 = 0.5
        a1_3 = 0.75
    elif abs(a2) < DOUBLE_EPS and abs(a1) > DOUBLE_INF:
        a = 0.0
        aa = 0.25
        a1_2 = 0.0
        a1_3 = 0.0
        a2_2 = 0.5
        a2_3 = 0.75
    else:
        assert abs(a1) > DOUBLE_EPS and abs(a2) > DOUBLE_EPS
        inv_a1 = 0.0 if abs(a1) > DOUBLE_INF else 1/a1
        inv_a2 = 0.0 if abs(a2) > DOUBLE_INF else 1/a2
        inv_a_demo = 1+4*inv_a1+4*inv_a2+12*inv_a1*inv_a2
        a = 1 / inv_a_demo
        aa = (1+inv_a1+inv_a2) / inv_a_demo
        a2_2 = (1+2*inv_a1) / inv_a_demo
        a2_3 = (1+3*inv_a1) / inv_a_demo
        a1_2 = (1+2*inv_a2) / inv_a_demo
        a1_3 = (1+3*inv_a2) / inv_a_demo

    # * the original rigid end formula
    # K[0,:] = np.array([12.,      sign*6*L, -12.,      sign*6*L])
    # K[1,:] = np.array([sign*6*L, 4*(L**2), sign*-6*L, 2*(L**2)])
    # K[2,:] = -K[0,:]
    # K[3,:] = np.array([sign*6*L, 2*(L**2), -sign*6*L, 4*(L**2)])
    # K *= (E*Iz/L**3)

    # * the modified formula for joint rotational stiffness
    # See [GMZ] p.394
    K[0,:] = np.array([12./L**2*aa, sign*6/L*a2_2, -12./L**2*aa, sign*6/L*a1_2])
    K[1,:] = np.array([sign*6/L*a2_2, 4*a2_3, sign*(-6/L)*a2_2, 2*a])
    K[2,:] = -K[0,:]
    K[3,:] = np.array([sign*6/L*a1_2, 2*a, sign*(-6/L)*a1_2, 4*a1_3])
    K *= E*Iz/L

    return K

def create_local_stiffness_matrix(L, A, Jx, Iy, Iz, E, mu, ct1=[np.inf]*3, ct2=[np.inf]*3, cr1=[np.inf]*3, cr2=[np.inf]*3):
    """complete 12x12 stiffness matrix for a bisymmetrical member.

        Since for small displacements the axial force effects, torsion,
        and bending about each axis are uncoupled, the influence coeff
        **relating** these effects are zero.
    
    Parameters
    ----------
    L : float
        element length
    A : float
        cross section area
    Jx : float
        torsional constant, unit: length unit^4
        In the case of circular, cylindrical shaft, it's equal to
        the polar moment of inertia of the cross section.
    Iy : float
        moment of inertia \int_A z^2 dA
    Iz : float
        moment of inertia \int_A y^2 dA
    E : float
        Young's modulus
    mu : float
        Poisson ratio
    ct1 : list of three float
        [Tx, Ty, Tz]
        translational spring stiffness at joint 1, set to 0 will release the corresponding dof
        default to np.inf (rigid joint)
    ct2 : list of three float
        [Tx, Ty, Tz]
        translational spring stiffness at joint 2, set to 0 will release the corresponding dof
        default to np.inf (rigid joint)
    cr1 : list of three float
        [Rx, Ry, Rz]
        rotational spring stiffness at joint 1, set to 0 will release the corresponding dof
        default to np.inf (rigid joint)
    cr2 : list of three float
        [Rx, Ry, Rz]
        rotational spring stiffness at joint 2, set to 0 will release the corresponding dof
        default to np.inf (rigid joint)
    
    Returns
    -------
    K : 12x12 numpy array
    """
    assert ct1[1] > DOUBLE_INF and ct2[1] > DOUBLE_INF, 'local y axis translational dof release not implemented!'
    assert ct1[2] > DOUBLE_INF and ct2[2] > DOUBLE_INF, 'local z axis translational dof release not implemented!'
    for i in range(3):
        assert abs(cr1[i]) < DOUBLE_EPS or abs(cr1[i]) > DOUBLE_INF, 'rotational stiffness has some bugs, not fully supported yet'
        assert abs(cr2[i]) < DOUBLE_EPS or abs(cr2[i]) > DOUBLE_INF, 'rotational stiffness has some bugs, not fully supported yet'

    G = mu2G(mu, E)
    # Fx1, Fx2 : u1, u2
    # 0, 6 : 0, 6
    axial_x_k = axial_stiffness_matrix(L, A, E, cr1=ct1[0], cr2=ct2[0])
    # Mx1, Mx2 : \theta_x1, \theta_x2
    # 3, 9 : 3, 9
    tor_x_k = torsional_stiffness_matrix(L, Jx, G, cr1=cr1[0], cr2=cr2[0])

    # Fy1, Mz1, Fy2, Mz2 : v1, \theta_z1, v2, \theta_z2
    # 1, 5, 7, 11 : 1, 5, 7, 11
    bend_z_k = bending_stiffness_matrix(L, E, Iz, axis=2, cr1=cr1[2], cr2=cr2[2])
    # Fz1, My1, Fz2, My2 : v1, \theta_z1, v2, \theta_z2
    # 2, 4, 8, 10 : 2, 4, 8, 10
    bend_y_k = bending_stiffness_matrix(L, E, Iz, axis=1, cr1=cr1[1], cr2=cr2[1])

    K = np.zeros([12,12])
    K[np.ix_([0,6], [0,6])] += axial_x_k
    K[np.ix_([3,9], [3,9])] += tor_x_k
    K[np.ix_([1,5,7,11], [1,5,7,11])] += bend_z_k
    K[np.ix_([2,4,8,10], [2,4,8,10])] += bend_y_k
    return K

def global2local_transf_matrix(end_vert_u, end_vert_v, rot_y2x=0.0):
    assert len(end_vert_u) == len(end_vert_v)
    assert len(end_vert_u) == 2 or len(end_vert_u) == 3
    dim = len(end_vert_u)
    L = norm(end_vert_u-end_vert_v)
    if L < 1e-6:
        raise ValueError('Bar length too small!: Node {} - Node {}'.format(end_vert_u, end_vert_v))

    # by convention, the new x axis is along the element's direction
    # directional cosine of the new x axis in the global world frame
    c_x = (end_vert_v[0] - end_vert_u[0])/L
    c_y = (end_vert_v[1] - end_vert_u[1])/L
    R = np.zeros([3,3])

    if rot_y2x != 0.0:
        raise NotImplementedError('rotation around local axis not implemented!')

    if 3 == dim:
        c_z = (end_vert_v[2] - end_vert_u[2]) / L
        # TODO rotaxis
        if abs(abs(c_z) - 1.0) < DOUBLE_EPS:
            # the element is parallel to global z axis
            # cross product is not defined, in this case
            # it's just a rotation about the global z axis
            # in x-y plane
            R[0, 2] = -c_z
            R[1, 1] = 1
            R[2, 0] = c_z
        else:
            # local x_axis = element's vector
            new_x = np.array([c_x, c_y, c_z])
            # local y axis = cross product with global z axis
            new_y = -np.cross(new_x, [0,0,1.0])
            new_y /= norm(new_y)
            new_z = np.cross(new_x, new_y)
            R[0, :] = new_x
            R[1, :] = new_y
            R[2, :] = new_z
        # R = R * rot_axis;
    elif 2 == dim:
        R[0,:] = [c_x, c_y, 0]
        R[1,:] = [-c_y, c_x, 0]
        R[2,2] = 1.0
        # TODO check rotaxis
    return R

def turn_diagblock(R3):
    """compile a 4x4 block-diagonal matrix

    Parameters
    ----------
    R3 : 3x3 numpy array

    Returns
    -------
    12x12 numpy array
    """
    R = np.zeros((12,12))
    for i in range(4):
        R[i*3:(i+1)*3, i*3:(i+1)*3] = R3
    return R

###############################################

def assemble_global_stiffness_matrix(k_list, nV, id_map, exist_e_ids=[]):
    """assemble element stiffness matrix into a sparse system stiffness matrix
    
    Parameters
    ----------
    k_list : list of np array
        a list of equal-sized global stiffness matrix
    nV : int
        number of vertices, used to compute total dof for sizing 
        the output sparse matrix
    id_map : np array
        (nEx(2xnode_dof)) idex matrix:
            M[element index] = [nodal dof index]
        e.g. for a single beam
            M[0, :] = [0, 1, 2, 3, 4, 5]

    Return
    ------
    Ksp : scipy.sparse.csc_matrix
        https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.sparse.csc_matrix.html#scipy.sparse.csc_matrix
    """
    node_dof = 6
    total_dof = node_dof * nV

    if len(exist_e_ids)==0:
        exist_e_ids = range(len(k_list))
    else:
        assert set(exist_e_ids) <= set(range(len(k_list)))

    row = []
    col = []
    data = []
    for e_id in exist_e_ids:
        Ke = k_list[e_id]
        assert Ke.shape == (node_dof*2, node_dof*2)
        for i in range(node_dof*2):
            for j in range(node_dof*2):
                if abs(Ke[i,j]) > DOUBLE_EPS:
                    row.append(id_map[e_id, i])
                    col.append(id_map[e_id, j])
                    data.append(Ke[i,j])
    Ksp = csc_matrix((data, (row, col)), shape=(total_dof, total_dof), dtype=float)
    return Ksp

###############################################

def compute_lumped_uniformly_distributed_load(w_G, R_LG, L):
    """Use fixed-end beam formula to lump uniformly distributed load 
    to equivalent nodal loads
    
    Parameters
    ----------
    w_G : np array of size (3,)
        uniformly distributed load vector in the global coordinate
        unit: force unit / length unit
    R_LG : np array of size (3,3)
        element local2global transformation matrix
    L : float
        element length
    """
    assert 3 == w_G.shape[0]
    lumped_L = np.zeros(12)
    # nodal force
    lumped_L[:3] = w_G*L/2
    lumped_L[6:9] = w_G*L/2
    # transform global load density to local density
    w_L = R_LG.dot(w_G)
    # node 0, 1 local moment
    M_L0 = np.array([0, -w_L[2]*(L**2)/12., w_L[1]*(L**2)/12.])
    M_L1 = -M_L0
    R_GL = R_LG.T
    lumped_L[3:6] = R_GL.dot(M_L0)
    lumped_L[9:12] = R_GL.dot(M_L1)
    return lumped_L

###############################################


import os
import warnings
from pyconmech.frame_analysis.io_base import Model, LoadCase
    
##################################################
# Stiffness Backend Template

class StiffnessBase(object):
    def __init__(self, model, verbose=False, output_json=False):
        self._nodes = model.nodes
        self._elements = model.elements
        self._supports = model.supports
        self._joints = model.joints
        self._materials = model.materials
        self._crosssecs = model.crosssecs

        self._verbose = verbose
        self._output_json = output_json
        self._trans_tol = None
        self._rot_tol = None

        if len(self._supports) == 0:
            raise RuntimeError('there needs to be at least one support (fixed) vertex in the model!')

    @classmethod
    def from_json(cls, json_file_path=None, verbose=False):
        """init class from a frame json file
        
        Parameters
        ----------
        json_file_path : [type], optional
            [description], by default None
        verbose : bool, optional
            [description], by default False
        
        Returns
        -------
        [type]
            [description]
        """
        return cls(Model.from_json(json_file_path, verbose=verbose), verbose=verbose)

    #######################
    # Load input

    def set_self_weight_load(self, include_self_weight=True, gravity_direction=[0,0,-1]):
        raise NotImplementedError()

    def set_load(self, nodal_forces):
        raise NotImplementedError()

    def set_uniformly_distributed_loads(self, element_load_density):
        raise NotImplementedError()

    def _parse_load_case_from_json(self, file_path):
        # TODO cpp engine only returns nodal load now
        raise NotImplementedError()
        # lc = LoadCase.from_json(file_path)
        # point_load, uniform_element_load, include_self_weight = \
        #     read_load_case_json(file_path)
        # return point_load, uniform_element_load, include_self_weight

    #######################
    # Output write

    def set_nodal_displacement_tol(self, transl_tol, rot_tol):
        raise NotImplementedError()

    def solve(self, exist_element_ids=[], if_cond_num=False):
        raise NotImplementedError()

    #######################
    # Output export

    def set_output_json_path(self, file_path, file_name):
    # .def("set_output_json_path", &conmech::stiffness_checker::Stiffness::setOutputJsonPath,
    #   py::arg("file_path"), py::arg("file_name"))
        raise NotImplementedError()

    def set_output_json(self, output_json=False):
    # .def("set_output_json", &conmech::stiffness_checker::Stiffness::setOutputJson,
    #   py::arg("output_json") = false)
        raise NotImplementedError()

    #######################
    # Get functions

    def get_frame_stat(self):
        raise NotImplementedError()

    def get_lumped_nodal_loads(self, existing_ids=[]):
        raise NotImplementedError()

    def get_gravity_nodal_loads(self, existing_ids=[]):
        raise NotImplementedError()

    def get_element_stiffness_matrices(self):
        raise NotImplementedError()

    def get_element_local2global_rot_matrices(self):
        raise NotImplementedError()

    def get_element2dof_id_map(self):
        raise NotImplementedError()

    def get_node2dof_id_map(self):
        raise NotImplementedError()

    def has_stored_result(self):
        raise NotImplementedError()

    def get_solved_results(self):
        raise NotImplementedError()

    def get_max_nodal_deformation(self):
        raise NotImplementedError()

    def get_nodal_deformation_tol(self):
        raise NotImplementedError()

    def get_original_shape(self):
        raise NotImplementedError()

    def get_deformed_shape(self, exagg_ratio=1.0, disc=10):
        raise NotImplementedError()

    # element attributes
    def get_element_crosssec(self, elem_id):
        e_tag = self._elements[elem_id].elem_tag
        # assert e_tag in self._crosssecs
        if e_tag in self._crosssecs:
            crosssec = self._crosssecs[e_tag]
        else:
            # TODO: default cross sec, if no [""] key is assigned
            warnings.warn('No cross section assigned for element tag |{}|, using the default tag'.format(e_tag))
            crosssec = self._crosssecs[None]
        return crosssec

    def get_element_material(self, elem_id):
        e_tag = self._elements[elem_id].elem_tag
        # assert e_tag in self._materials
        if e_tag in self._materials:
            mat = self._materials[e_tag]
        else:
            # TODO: default material, if no [""] key is assigned
            warnings.warn('No material assigned for element tag |{}|, using the default tag'.format(e_tag))
            mat = self._materials[None]
        return mat


import os, sys

assert 'ironpython' in sys.version.lower() and os.name == 'nt', 'Karamba only available for IronPython on Windows now.'

THIS_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(THIS_DIR))
from io_base import Node, Element, Support, Joint, CrossSec, Material, Model, LoadCase, \
    PointLoad, UniformlyDistLoad, GravityLoad

###############################################

# https://ironpython.net/documentation/dotnet/dotnet.html
import clr
clr.AddReferenceToFileAndPath("C:\Program Files\Rhino 6\Plug-ins\Karamba.gha")
clr.AddReferenceToFileAndPath("C:\Program Files\Rhino 6\Plug-ins\KarambaCommon.dll")

import feb # Karamba's C++ library (undocumented in the API)
import Karamba
import Karamba.Models.Model
from Karamba.Geometry import Point3, Line3, Vector3, Plane3
import Karamba.Nodes
import Karamba.Elements
import Karamba.CrossSections
import Karamba.Materials
import Karamba.Supports
import Karamba.Loads
from Karamba.Utilities import MessageLogger, UnitsConversionFactories
import KarambaCommon
#
import System
from System import GC
from System.Collections.Generic import List
from System.Drawing import Color

####################################

class KarambaNode(Node):
    @classmethod
    def from_karamba(cls, knode, is_grounded=False):
        point = [knode.pos.X, knode.pos.Y, knode.pos.Z]
        return cls(point, knode.ind, is_grounded)
    
    def to_karamba(self, type='Point3'):
        if type == 'Point3':
            return Point3(*self.point)
        elif type == 'Node':
            return Karamba.Nodes.Node(self.node_ind, Point3(*self.point))
        else:
            raise ValueError

class KarambaElement(Element):
    @classmethod
    def from_karamba(cls, kelement):
        return cls(list(kelement.node_inds), kelement.ind, elem_tag=kelement.id, 
            bending_stiff=kelement.bending_stiff)

    def to_karamba(self, type='builderbeam'):
        if type == 'builderbeam':
            bb = Karamba.Elements.BuilderBeam(*self.end_node_inds)
            bb.id = self.elem_tag
            bb.bending_stiff = self.bending_stiff
            return bb
        else:
            raise NotImplementedError

class KarambaSupport(Support):
    @classmethod
    def from_karamba(cls, ksupport):
        return cls(list(ksupport.Condition), ksupport.node_ind)

    def to_karamba(self):
        return Karamba.Supports.Support(self.node_ind, List[bool](self.condition), Plane3())

class KarambaJoint(Joint):
    @classmethod
    def from_karamba(cls, kjoint):
        assert kjoint.dofs == 6
        # condition is If null there is no joint for the corresponding DOF. 
        return cls(list(kjoint.c_condition(0)) + list(kjoint.c_condition(1)), list(kjoint.elemIds))

    def to_karamba(self):
        jt = Karamba.Joints.Joint()
        jt.elemIds = List[str](self.elem_tags)
        jt.c = System.Array[System.Nullable[float]](self.c_conditions)
        return jt

class KarambaCrossSec(CrossSec):
    @classmethod
    def from_karamba(cls, kc):
        return cls(kc.A, kc.Ipp, kc.Iyy, kc.Izz, elem_tags=list(kc.elemIds), family=kc.family, name=kc.name)

    def to_karamba(self):
        beam_mod = Karamba.CrossSections.CroSec_BeamModifier()
        beam_mod.A = self.A
        beam_mod.Ipp = self.Jx
        beam_mod.Iyy = self.Iy
        beam_mod.Izz = self.Iz
        beam_mod.elemIds = List[str]([tag for tag in self.elem_tags if tag is not None])
        # beam_mod.clearElemIds()
        # for tag in self.elem_tags:
        #     if tag is not None:
        #         beam_mod.AddElemId(tag)
        beam_mod.family = self.family
        beam_mod.name = self.name
        return beam_mod

class KarambaMaterialIsotropic(Material):
    @classmethod
    def from_karamba(cls, km):
        assert km.typeName() == 'ISO', 'Input should be a Isotropic material!'
        return cls(km.E(), km.G12(), km.fy(), km.gamma(), elem_tags=list(km.elemIds), 
            family=km.family, name=km.name, type_name=km.typeName(), G3=km.G3())

    def to_karamba(self):
        steel_alphaT = 1e-5 #1/deg
        km = Karamba.Materials.FemMaterial_Isotrop(
	        self.family, self.name, self.E, self.G12, self.G3, self.density, self.fy, steel_alphaT, None) #Color.FromName("SlateBlue")
        for tag in self.elem_tags:
            if tag is not None:
                km.AddBeamId(tag)
        return km

class KarambaPointLoad(PointLoad):
    @classmethod
    def from_karamba(cls, kf):
        return cls([kf.force.X, kf.force.Y, kf.force.Z], 
                   [kf.moment.X, kf.moment.Y, kf.moment.Z], kf.node_ind, kf.loadcase)

    def to_karamba(self, corotate=False):
        # corotate = true the pointload corotates with the node
        pl = Karamba.Loads.PointLoad(self.node_ind, Vector3(*self.force), Vector3(*self.moment), corotate)
        pl.loadcase = self.loadcase
        return pl

class KarambaUniformlyDistLoad(UniformlyDistLoad):
    @classmethod
    def from_karamba(cls, kf):
        return cls([kf.Q.X, kf.Q.Y, kf.Q.Z], 
                   [kf.Load.X, kf.Load.Y, kf.Load.Z], kf.beamIds, loadcase=kf.loadcase)

    def to_karamba(self):
        # Karamba.Loads.LoadOrientation.global
        ul = Karamba.Loads.UniformlyDistLoad(List[str](self.elem_tags), Vector3(*self.q), 0, self.loadcase)
        return ul

class KarambaGravityLoad(GravityLoad):
    @classmethod
    def from_karamba(cls, kf):
        return cls([kf.force.X, kf.force.Y, kf.force.Z], kf.loadcase)

    def to_karamba(self):
        return Karamba.Loads.GravityLoad(Vector3(*self.force), self.loadcase)

class KarambaLoadCase(LoadCase):
    @classmethod
    def from_loadcase(cls, loadcase):
        kpoint_loads = [KarambaPointLoad.from_data(pl.to_data()) for pl in loadcase.point_loads]
        kelem_loads = [KarambaUniformlyDistLoad.from_data(pl.to_data()) for pl in loadcase.uniform_element_loads]
        kgravity_load = None
        if loadcase.gravity_load is not None:
            kgravity_load = KarambaGravityLoad.from_data(loadcase.gravity_load.to_data())
        return cls(kpoint_loads, kelem_loads, kgravity_load)

    @classmethod
    def from_karamba(cls, km, lc=0):
        assert lc < km.numLC
        point_loads = []
        for pl in km.ploads:
            if pl.loadcase == lc:
                point_loads.append(KarambaPointLoad.from_karamba(pl))

        uniform_element_loads = []
        for el in km.eloads:
            if el.loadcase == lc and isinstance(el, Karamba.Loads.UniformlyDistLoad):
                uniform_element_loads.append(KarambaUniformlyDistLoad.from_karamba(el))

        gravity_load = None
        k_grav = clr.Reference[Karamba.Loads.GravityLoad]()
        km.gravities.TryGetValue(lc, k_grav)
        if k_grav.Value is not None:
            gravity_load = KarambaGravityLoad.from_karamba(k_grav.Value)
        
        return cls(point_loads, uniform_element_loads, gravity_load)

    def to_karamba(self):
        in_loads = List[Karamba.Loads.Load]()
        in_loads.AddRange([pl.to_karamba() for pl in self.point_loads])
        in_loads.AddRange([pl.to_karamba() for pl in self.uniform_element_loads])
        if self.gravity_load is not None:
            in_loads.Add(self.gravity_load.to_karamba())
        return in_loads

####################################

class KarambaModel(Model):
    @classmethod
    def from_model(cls, model):
        knodes = [KarambaNode.from_data(n.to_data()) for n in model.nodes]
        kelements = [KarambaElement.from_data(e.to_data()) for e in model.elements]
        ksupports = [KarambaSupport.from_data(s.to_data()) for s in model.supports.values()]
        kjoints = [KarambaJoint.from_data(j.to_data()) for j in model.joints.values()]
        kmaterials = [KarambaMaterialIsotropic.from_data(m.to_data()) for m in model.materials.values()]
        kcrosssecs = [KarambaCrossSec.from_data(cs.to_data()) for cs in model.crosssecs.values()]
        return cls(knodes, kelements, ksupports, kjoints, kmaterials, kcrosssecs, 
            unit=model.unit, model_name=model.model_name)

    @classmethod
    def from_data(cls, data, verbose=False):
        model = super(KarambaModel, cls).from_data(data, verbose)
        return cls.from_model(model)

    @classmethod
    def from_karamba(cls, km, grounded_supports=None, model_name='', limit_distance=1e-6):
        knodes = [KarambaNode.from_karamba(kn) for kn in km.nodes]
        kelements = [KarambaElement.from_karamba(e) for e in km.elems]
        ksupports = [KarambaSupport.from_karamba(s) for s in km.supports]
        kjoints = [KarambaJoint.from_karamba(j) for j in km.joints]
        kmaterials = [KarambaMaterialIsotropic.from_karamba(m) for m in km.materials]
        kcrosssecs = [KarambaCrossSec.from_karamba(cs) for cs in km.crosecs]

        if grounded_supports is not None:
            for k_supp in km.supports:
                for k_g in grounded_supports:
                    if k_supp.position.DistanceTo(k_g.position) < limit_distance:
                      knodes[k_supp.node_ind].is_grounded = True

        return cls(knodes, kelements, ksupports, kjoints, kmaterials, kcrosssecs, 
            unit="meter", model_name=model_name)

    def to_karamba(self, loadcase, limit_dist=1e-6):
        mass = clr.Reference[float]()
        cog = clr.Reference[Point3]()
        warning_flag = clr.Reference[bool]()
        info = clr.Reference[str]()
        msg = clr.Reference[str]()
        kmodel = clr.Reference[Karamba.Models.Model]()

        in_points = List[Point3]([n.to_karamba(type='Point3') for n in self.nodes])
        in_elems = List[Karamba.Elements.BuilderElement]([e.to_karamba(type='builderbeam') for e in self.elements])
        for e in in_elems:
            eid = None if e.id not in self.crosssecs else e.id
            e.crosec = self.crosssecs[eid].to_karamba()
        in_supports = List[Karamba.Supports.Support]([s.to_karamba() for _, s in self.supports.items()])

        in_crosecs = List[Karamba.CrossSections.CroSec]([cs.to_karamba() for _, cs in self.crosssecs.items()])
        in_materials = List[Karamba.Materials.FemMaterial]([m.to_karamba() for _, m in self.materials.items()])
        in_joints = List[Karamba.Joints.Joint]([j.to_karamba() for _, j in self.joints.items()])

        kloadcase = KarambaLoadCase.from_loadcase(loadcase)
        in_loads = kloadcase.to_karamba()

        Karamba.Models.AssembleModel.solve(
        	in_points,
        	in_elems,
        	in_supports,
        	in_loads,
        	in_crosecs,
        	in_materials,
        	List[Karamba.Elements.ElemSet]([]),
        	in_joints,
        	limit_dist,
            # out
        	kmodel, info, mass, cog, msg, warning_flag)

        if warning_flag.Value:
            print('info {} | msg {}'.format(info.Value, msg.Value))

        # # calculate Th.I response
        # max_disp = clr.Reference[List[float]]()
        # out_g = clr.Reference[List[float]]()
        # out_comp = clr.Reference[List[float]]()
        # message = clr.Reference[str]()
        # model = k3d.Algorithms.AnalyzeThI(model, max_disp, out_g, out_comp, message);

        # ucf = UnitsConversionFactories.Conv();
        # cm = ucf.cm();
        # print("max disp: {} {}".format(cm.toUnit(max_disp.Value[0]), cm.unitB))

        return kmodel.Value


import warnings
import numpy as np
import scipy
import scipy.sparse.linalg as SPLA
from scipy.sparse import csc_matrix
import numpy.linalg as LA
from collections import defaultdict
from numpy.linalg import norm
from termcolor import cprint

from pyconmech.frame_analysis.stiffness_base import StiffnessBase
from pyconmech.frame_analysis.numpy_stiffness_assembly import DOUBLE_EPS, DOUBLE_INF, \
    create_local_stiffness_matrix, global2local_transf_matrix, turn_diagblock, assemble_global_stiffness_matrix, \
    compute_lumped_uniformly_distributed_load

DEFAULT_GRAVITY_DIRECTION = np.array([0.,0.,-1.])

############################################

# internal calculation unit
LENGTH_UNIT = 'm' # meter
ROT_ANGLE_UNIT = 'rad'
FORCE_UNIT = 'kN'

# derived units
MOMENT_UNIT = FORCE_UNIT + "*" + ROT_ANGLE_UNIT
AREA_UNIT = LENGTH_UNIT + "2"
AREA_INERTIA_UNIT = LENGTH_UNIT + "4"
PRESSURE_UNIT = FORCE_UNIT + "/" + LENGTH_UNIT + "2"; # "kN/m^2" modulus
DENSITY_UNIT  = FORCE_UNIT + "/" + LENGTH_UNIT + "3"; # "kN/m^3"

############################################

class NumpyStiffness(StiffnessBase):
    def __init__(self, model, verbose=False, output_json=False):
        super(NumpyStiffness, self).__init__(model, verbose, output_json)

        # default displacement tolerance
        self._trans_tol = 1e-3
        self._rot_tol = np.pi/180 * 3
        # gravity settings
        self._include_self_weight_load = False
        # self._full_node_dof = 6
        # self._node_dof = -1
        # internal states
        self._is_init = False
        # element id to dof map
        self._id_map = None
        # node id to dof map
        self._v_id_map = None
        # node id -> element id
        self._node_neighbors = self.get_node_neighbors()
        # elem tag -> elem ind
        self._e_ind_from_tag = self.get_elem_ind_from_elem_tag()

        # (nE x (12x12)) list of element stiffness matrix
        self._element_k_list = []
        # (nE x (12x12)) block-diagnal global2local coordinate transformation matrix
        self._rot_list = []
        # (_ x (6)) dict of lumped element nodal load in global coordinate
        self._element_lumped_nload_list = {}
        # (nE x (12)) list of lumped GRAVITY element nodal load in global coordinate
        self._element_gravity_nload_list = {}
        # (nVx6), np array
        self._nodal_load = None
        # stored computation results
        self._has_stored_deformation = False
        self._stored_existing_ids = []
        self._stored_nD = None # nodal displacement, (nExistNode x 7) np array
        self._stored_eR = None # element reaction, (nExistElement x 13) np array
        self._stored_fR = None # fixities reaction, (nExistFixities x 6) np array
        self._stored_compliance = None
        # output settings
        self._output_json_file_name = ""
        self._output_json_file_path = ""
        # used in non-frame model
        # translational dof id in [0, ..., 2*node_dof-1]
        self._xyz_dof_id = None
        # elemental reaction dof id in [0, ..., 2*node_dof-1]
        self._e_react_dof_id = None

        self._init()

    @property
    def nV(self):
        """number of vertices in the full model
        """
        return len(self._nodes)

    @property
    def nE(self):
        """number of elements in the full model
        """
        return len(self._elements)

    #######################
    # Load input
    def set_self_weight_load(self, include_self_weight=True, gravity_direction=DEFAULT_GRAVITY_DIRECTION):
        self._include_self_weight_load = include_self_weight
        if include_self_weight:
            assert len(gravity_direction) == 3
            gravity_direction = np.array(gravity_direction)
            self_weight_load_density = {}
            for i in range(self.nE):
                crosssec = self.get_element_crosssec(i)
                mat = self.get_element_material(i)
                q_sw = mat.density * crosssec.A
                w_G = gravity_direction * q_sw
                self_weight_load_density[i] = w_G
            self._element_gravity_nload_list = \
                self.compute_element_uniformly_distributed_lumped_load(self_weight_load_density)

    def set_uniformly_distributed_loads(self, unif_dist_loads):
        if unif_dist_loads is not None:
            e_load_density_from_ind = {}
            for e_tag, uload in unif_dist_loads.items():
                for e_ind in self._e_ind_from_tag[e_tag]:
                    e_load_density_from_ind[e_ind] = uload.q
            self._element_lumped_nload_list = \
                self.compute_element_uniformly_distributed_lumped_load(e_load_density_from_ind)
        else:
            # reset
            self._element_lumped_nload_list = {}

    def set_load(self, nodal_forces):
        """[summary]
        
        Parameters
        ----------
        nodal_forces : dict
            {node id : PointLoad}
        """
        if nodal_forces is not None:
            for v_id, pf in nodal_forces.items():
                self._nodal_load[self._v_id_map[v_id,:]] = np.array(list(pf.force) + list(pf.moment))
        else:
            # reset
            self._nodal_load = np.zeros(self.nV*6)

    #######################
    # Compute functions
    def set_nodal_displacement_tol(self, transl_tol, rot_tol):
        # in meter, rad
        self._trans_tol = transl_tol
        self._rot_tol = rot_tol

    def compute_dof_permutation(self, exist_element_ids=[]):
        """a row permuatation matrix (multiply left)
        """
        if len(exist_element_ids)==0:
            exist_element_ids = range(self.nE)

        total_dof = 6*self.nV
        # assume all dof does not exist (-1)
        # 0: free, -1: not exist, 1: fixed
        dof_stat = np.ones(total_dof, dtype=int)*-1

        # existing id set
        exist_node_set = set()
        for e in exist_element_ids:
            exist_node_set.update(self._elements[e].end_node_inds)
            # turn the existing node's dof status to free (0)
            dof_stat[self._id_map[e]] = np.zeros(12)

        # assemble the list of fixed dof fixed list, 1x(nFv)
        if len(set(self._supports.keys()) & exist_node_set) == 0:
            if self._verbose : print('Structure floating: no fixed node exists in the partial structure.')
            return False
        n_exist_fixed_nodes = 0
        n_fixed_dof = 0
        for fv_id, support in self._supports.items():
            if fv_id in exist_node_set:
                dof_stat[self._v_id_map[fv_id,:]] = np.array(support.condition, dtype=int)
                # if all existing element bending=False, fix rotational dof
                # to avoid singularity issue
                # TODO partial assembled state?
                # if all([not self._elements[e_id].bending_stiff or e_id not in exist_element_ids for e_id in self._node_neighbors[fv_id]]):
                if all([not self._elements[e_id].bending_stiff for e_id in self._node_neighbors[fv_id]]):
                    # print('pinned dof because of inactive bending')
                    dof_stat[self._v_id_map[fv_id,3:]] = [1,1,1]
                n_exist_fixed_nodes+=1
                assert all([s >= 0 for s in dof_stat[self._v_id_map[fv_id,:]]])
                n_fixed_dof += sum(dof_stat[self._v_id_map[fv_id,:]])

        if n_fixed_dof == 0:
            if self._verbose:
                print('Not stable: At least one node needs to be fixed in the considered substructure!')
            return False
        n_free_dof = 6*len(exist_node_set)-n_fixed_dof

        # compute permutation map
        free_tail = 0
        fix_tail = n_free_dof
        nonexist_tail = 6*len(exist_node_set)
        id_map_RO = np.array(range(total_dof))
        for i in range(total_dof):
            if 0 == dof_stat[i]:
                id_map_RO[free_tail] = i
                free_tail += 1
            elif 1 == dof_stat[i]:
                id_map_RO[fix_tail] = i
                fix_tail += 1
            elif -1 == dof_stat[i]:
                id_map_RO[nonexist_tail] = i
                nonexist_tail += 1
            else:
                raise ValueError

        # a row permuatation matrix (multiply left)
        perm_data = []
        perm_row = []
        perm_col= []
        for i in range(total_dof):
            perm_row.append(i)
            perm_col.append(id_map_RO[i])
            perm_data.append(1)
        Perm = csc_matrix((perm_data, (perm_row, perm_col)), shape=(total_dof, total_dof))
        return Perm, (exist_node_set, n_free_dof, n_fixed_dof, n_exist_fixed_nodes)

    def compute_full_stiffness_matrix(self, exist_element_ids=[]):
        if len(exist_element_ids)==0:
            exist_element_ids = range(self.nE)
        K_full = assemble_global_stiffness_matrix(self._element_k_list, \
            self.nV, self._id_map, exist_element_ids)
        return K_full

    def solve(self, exist_element_ids=[], if_cond_num=False):
        # assume full structure if exist_element_ids is []
        if len(exist_element_ids)==0:
            exist_element_ids = range(self.nE)

        total_dof = 6*self.nV
        
        Perm, (exist_node_set, n_free_dof, n_fixed_dof, n_exist_fixed_nodes) = self.compute_dof_permutation(exist_element_ids)
        PermT = Perm.T

        # permute the full stiffness matrix & carve the needed portion out
        K_full_ = self.compute_full_stiffness_matrix(exist_element_ids)
        K_perm = (Perm.dot(K_full_)).dot(PermT)
        K_mm = K_perm[0:n_free_dof, 0:n_free_dof]
        K_fm = K_perm[n_free_dof:n_free_dof+n_fixed_dof, 0:n_free_dof]

        # TODO make the load vector P sparse
        # assemble load vector
        nodal_load_P_tmp = np.copy(self._nodal_load)
        if len(self._element_lumped_nload_list) > 0:
            # if element unif load is specified
            element_lumped_load = self._create_uniformly_distributed_lumped_load(exist_element_ids)
            nodal_load_P_tmp += element_lumped_load
        if self._include_self_weight_load:
            load_sw = self._create_self_weight_load(exist_element_ids)
            nodal_load_P_tmp += load_sw
        # nodal_load_P_tmp *= -1

        P_m = np.zeros(n_free_dof)
        P_f = np.zeros(n_fixed_dof)
        # reset results
        U_m = np.zeros(n_free_dof)
        self._stored_compliance = 0.0

        # solve
        if norm(nodal_load_P_tmp) > DOUBLE_EPS:
            P_perm = Perm.dot(nodal_load_P_tmp) 
            P_m = P_perm[0:n_free_dof]
            P_f = P_perm[n_free_dof:n_free_dof+n_fixed_dof]

            # TODO Karamba warning when plugging PLA's properties in
            # Material #0 : For isotropic materials 
            # G must be larger than E/3 and smaller than E/2. 
            # If this condition is not fulfilled the material behaves very strange. 
            # It may lead to an unstable structure. May cause errors in exported models.

            # * preconditioning K_mm by K_mm.diagonal()
            # Ax = b
            # (D1 A D2) x_ = (D1) b
            # x = D2 x_
            # take D1 = D2 = [ 1/sqrt(a_{ii}) ]
            if all([abs(a) > DOUBLE_EPS for a in K_mm.diagonal()]):
                # P = diag(1/a_ii) = E_inv.T * E_inv
                # P_inv = diag(a_ii)
                # P_inv = E * E.T
                E_inv_diag = np.array([1/np.sqrt(a) for a in K_mm.diagonal()])
                E_inv = scipy.sparse.diags(E_inv_diag, format='csc')
            else:
                E_inv = scipy.sparse.eye(K_mm.shape[0], format='csc')
            K_mm_precond = (E_inv.dot(K_mm)).dot(E_inv)

            # TODO serialize and check PD and conditioning
            # https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.sparse.linalg.SuperLU.html#scipy.sparse.linalg.SuperLU
            # * sparse LU decomposition
            # See 'Symmetric or diagonally dominant problems' in: https://portal.nersc.gov/project/sparse/superlu/faq.html
            # tell SuperLU to enter the Symmetric Mode, see example:
            # https://github.com/xiaoyeli/superlu/blob/a3d5233770f0caad4bc4578b46d3b26af99e9c19/EXAMPLE/dlinsol1.c#L60
            try:
                K_LU = SPLA.splu(K_mm_precond, diag_pivot_thresh=0, options=dict(SymmetricMode=True)) 
            except RuntimeError as exception:
                if self._verbose:
                    print ('Sparse solve error: {}'.format(exception))
                
                    eigvals = SPLA.eigsh(K_mm_precond, k=int(K_mm_precond.shape[0]/2.0), which='SM', return_eigenvectors=False) 
                    # eigvals, eigvecs = scipy.linalg.eigh(K_mm_precond.toarray())
                    n_zero_eigs = 0
                    for _, eigval in enumerate(eigvals):
                        if abs(eigval) < DOUBLE_EPS:
                            n_zero_eigs += 1
                            # print(norm(eigvecs[eig_i]))
                    if n_zero_eigs > 0:
                        print('There are {}/{} rigid body modes in the system. This means some parts can move freely without causing deformation.'.format(
                            n_zero_eigs, K_mm.shape[0]))
                    #  Try to use the 'Eigen Modes'-component and activate the display of local coordinate axes: The first eigen-mode will be the rigid body motion.
                    #  If this does not help, check whether you have a pinned support directly attached to a hinge. A hinge introduces an extra node which may cause the problem.
                    #  When analyzing a flat shell structure one has to lock the rotation perpendicular to the plate in at least one node.
                return False

            # * positive definiteness can be checked by
            #   - LDLt, check if D is all positive
            #   - two ends of the spectrum are positive
            # TODO which one is more stable? faster?

            if not (K_LU.U.diagonal()>-DOUBLE_EPS).all():
                # https://gist.github.com/omitakahiro/c49e5168d04438c5b20c921b928f1f5d
                # https://docs.scipy.org/doc/scipy-1.4.1/reference/generated/scipy.sparse.linalg.SuperLU.html#scipy.sparse.linalg.SuperLU
                # Note: the permutation is not checker here
                print('perm: ', (K_LU.perm_r == np.arange(K_mm_precond.shape[0])).all() )
                # when a diagonal entry is smaller than the threshold, the code will still choose an off-diagonal pivot.
                # That is, the row permutation P_r may not be Identity.
                if self._verbose : 
                    print('ERROR: Stiffness Solver fail! Stiffness matrix not Positive Definite, the sub-structure might contain mechanism.')
                return False

            # ? Method 1: direct solve
            # U_m_pred = SPLA.spsolve(K_mm_precond, E_inv.dot(P_m))

            # ? Method 2: CG variant
            # https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.linalg.gmres.html#scipy.sparse.linalg.gmres
            # U_m, exit_code = SPLA.gmres(K_mm_precond, P_m, tol=1e-7)
            # if exit_code != 0:
            #     if self._verbose:
            #         if exit_code > 0:
            #             print('# iterations: {} - {}'.format(exit_code, 'convergence to tolerance not achieved'))
            #         elif exit_code < 0:
            #             print('exit_code : {} - {}'.format(exit_code, 'illegal input or breakdown'))
            #     return False

            # ? Method 3: SuperLU
            U_m_pred = K_LU.solve(E_inv.dot(P_m))

            # undo the precondition
            U_m = E_inv * U_m_pred

            # gathering results
            self._stored_compliance = 0.5*U_m.dot(P_m)
        else:
            if self._verbose:
                cprint('No external force is applied to the structure.', 'yellow')
            # warnings.warn('No external force is applied to the structure.')

        # reaction
        R = np.zeros(total_dof)
        R[n_free_dof:n_free_dof+n_fixed_dof] = K_fm.dot(U_m) - P_f
        R = PermT.dot(R)
        # displacement
        U = np.zeros(total_dof)
        U[0:n_free_dof] = U_m
        U = PermT.dot(U)

        # raw result conversion        
        self._stored_nD = np.zeros((len(exist_node_set), 1+6), dtype=float)
        for i, v in enumerate(list(exist_node_set)):
            self._stored_nD[i, :] = np.hstack([[v], U[self._v_id_map[v,:]]])

        self._stored_fR = np.zeros((n_exist_fixed_nodes, 1+6), dtype=float)
        cnt = 0
        for fv_id in self._supports:
            if fv_id in exist_node_set:
                self._stored_fR[cnt, :] = np.hstack([[fv_id], R[fv_id*6:(fv_id+1)*6]])
                cnt += 1

        self._stored_eR = np.zeros((len(exist_element_ids), 1+12), dtype=float)
        for i, e in enumerate(exist_element_ids):
            Ue = U[self._id_map[e]]
            eRF = self._rot_list[e].dot(self._element_k_list[e].dot(Ue))
            self._stored_eR[i, :] = np.hstack([[e], eRF])

        self._stored_existing_ids = np.copy(exist_element_ids)
        self._has_stored_deformation = True

        return self.check_stiffness_criteria(self._stored_nD, \
            self._stored_fR, self._stored_eR)

    #######################
    # Public compute functions that are non-state-changing
    def compute_element_uniformly_distributed_lumped_load(self, e_load_density_from_ind):
        """Use fixed-end beam formula to lump uniformly distributed load 
        to equivalent nodal loads
        
        Parameters
        ----------
        element_load_density : dict: int->np array of size (3,)
            load density dict that maps element id to load density vector of size 3
        
        Returns
        -------
        element_lumped_nload_list : dict: e_id : np array of size (3,)
        """
        # refer [MSA McGuire et al.] P111
        # Loads between nodal points
        element_lumped_nload_list = {i : np.zeros(12) for i in range(self.nE)}
        # element_lumped_nload_list = {}
        for e_ind, w_G in e_load_density_from_ind.items():
            end_u_id, end_v_id = self._elements[e_ind].end_node_inds
            end_u = np.array(self._nodes[end_u_id].point)
            end_v = np.array(self._nodes[end_v_id].point)
            L = norm(end_u-end_v)
            try:
                R3 = global2local_transf_matrix(end_u, end_v)
            except ValueError as err:
                raise ValueError('E#{} - {}'.format(e_ind, err))
            lumped_L = compute_lumped_uniformly_distributed_load(np.array(w_G), R3, L)
            assert lumped_L.shape[0] == 12
            element_lumped_nload_list[e_ind] = lumped_L
        return element_lumped_nload_list

    def compute_nodal_loads(self, nodal_forces):
        ext_load = np.zeros(6*self.nV)
        for v, nF in nodal_forces.items():
            ext_load[v*6:(v+1)*6] = nF
        return ext_load
    
    def check_stiffness_criteria(self, node_displ, fixities_reaction, element_reaction, ord=None):
        # user can overwrite this function
        nD_trans_max, nD_rot_max, _, _ = self.get_max_nodal_deformation(ord=ord)
        return nD_trans_max < self._trans_tol

    #######################
    # Get functions - Main results
    def has_stored_result(self):
        return self._has_stored_deformation

    def get_solved_results(self):
        assert self.has_stored_result()
        success = self.check_stiffness_criteria(self._stored_nD, self._stored_fR, self._stored_eR)
        return (success, self._stored_nD, self._stored_fR, self._stored_eR)

    def get_compliance(self):
        assert self.has_stored_result()
        return self._stored_compliance

    def get_max_nodal_deformation(self, ord=None):
        assert self.has_stored_result()
        # print(self._stored_nD)
        # print(LA.norm(self._stored_nD[:,1:4], ord=ord, axis=1))
        nD_trans_max = LA.norm(self._stored_nD[:,1:4], ord=ord, axis=1)
        nD_rot_max = LA.norm(self._stored_nD[:,4:7], ord=ord, axis=1)
        return (np.max(nD_trans_max), np.max(nD_rot_max),\
                np.argmax(nD_trans_max), np.argmax(nD_rot_max))

    # settings getters
    def get_nodal_deformation_tol(self):
        return (self._trans_tol, self._rot_tol)

    def get_frame_stat(self):
        return (self.nV, self.nE)

    #######################
    # Get functions - procedural data
    def get_lumped_nodal_loads(self, existing_ids=[]):
        """packing all applied nodal load and return the load vector
        
        Parameters
        ----------
        existing_ids : list of int, optional
            existing element's indices, by default []
        
        Returns
        -------
        np array of size (nV * 6)

        """
        if len(existing_ids) == 0:
            existing_ids = range(self.nE)
        P = np.copy(self._nodal_load)
        if len(self._element_lumped_nload_list) > 0:
            # if element unif load is specified
            P += self._create_uniformly_distributed_lumped_load(existing_ids)
        if self._include_self_weight_load:
            P += self.get_gravity_nodal_loads(existing_ids)
        return P

    def get_gravity_nodal_loads(self, existing_ids=[]):
        if len(existing_ids)==0:
            existing_ids = range(self.nE)
        return self._create_self_weight_load(existing_ids)

    def get_element_stiffness_matrices(self):
        return self._element_k_list

    def get_element_local2global_rot_matrices(self):
        return self._rot_list

    def get_element2dof_id_map(self):
        return self._id_map

    def get_node2dof_id_map(self):
        return self._v_id_map

    def get_node_neighbors(self):
        """Return all nodes' connected element ids
        
        Returns
        -------
        node_neighbors : dict
            {node_id : set(connected element ids)}
        """
        node_neighbors = defaultdict(set)
        for e in self._elements:
            n1, n2 = e.end_node_inds
            node_neighbors[n1].add(e.elem_ind)
            node_neighbors[n2].add(e.elem_ind)
        return node_neighbors

    def get_elem_ind_from_elem_tag(self):
        e_ind_from_tag = {}
        for e in self._elements:
            if e.elem_tag not in e_ind_from_tag:
                e_ind_from_tag[e.elem_tag] = []
            e_ind_from_tag[e.elem_tag].append(e.elem_ind)
        return e_ind_from_tag

    ################################
    # Visualizations
    def get_original_shape(self):
        raise NotImplementedError()

    def get_deformed_shape(self, exagg_ratio=1.0, disc=10):
        raise NotImplementedError()

    #######################
    # Output export
    def set_output_json_path(self, file_path, file_name):
        raise NotImplementedError()

    def set_output_json(self, output_json=False):
        raise NotImplementedError()

    ################################
    # private functions - init & precompute
    def _init(self, verbose=False):
        # else:
            # self._node_dof = 6
            # self._xyz_dof_id = np.array(range(2*self._node_dof))
            # self._e_react_dof_id = np.array(range(2*self._node_dof))
        self._precompute_element_stiffness_matrices()
        # node id-to-dof map
        self._v_id_map = np.zeros((self.nV, 6), dtype=int)
        for i in range(self.nV):
            self._v_id_map[i, :] = np.array(range(i*6, (i+1)*6))
        # element id-to-dof map
        self._id_map = np.zeros((self.nE, 12), dtype=int)
        for i in range(self.nE):
            end_u_id, end_v_id = self._elements[i].end_node_inds
            self._id_map[i, :] = np.hstack([self._v_id_map[end_u_id,:], self._v_id_map[end_v_id,:]])
        # loads
        self._nodal_load = np.zeros(self.nV*6)
        if self._include_self_weight_load:
            self.set_self_weight_load(True, DEFAULT_GRAVITY_DIRECTION)
        self._is_init = True
        if verbose:
            print('NpStiffness init done.')

    def _precompute_element_stiffness_matrices(self):
        self._element_k_list = [np.zeros((12,12)) for i in range(self.nE)]
        self._rot_list = [np.zeros((12,12)) for i in range(self.nE)]
        for i in range(self.nE):
            element = self._elements[i]
            e_tag = element.elem_tag
            end_u_id, end_v_id = element.end_node_inds
            end_u = np.array(self._nodes[end_u_id].point)
            end_v = np.array(self._nodes[end_v_id].point)
            L = norm(end_u-end_v)

            cr1 = [np.inf for _ in range(3)]
            cr2 = [np.inf for _ in range(3)]
            if e_tag in self._joints:
                assert all([c is None or c == np.inf for c in self._joints[e_tag].c_conditions[0:3]]) and \
                    all([c is None or c == np.inf for c in self._joints[e_tag].c_conditions[6:9]]), "We do not support translational joint now!"
                cr1 = [np.inf if c is None else c for c in self._joints[e_tag].c_conditions[3:6]]
                cr2 = [np.inf if c is None else c for c in self._joints[e_tag].c_conditions[9:12]]
            if not element.bending_stiff:
                cr1 = [0.0 for _ in range(3)]
                cr2 = [0.0 for _ in range(3)]
            try:
                R3 = global2local_transf_matrix(end_u, end_v)
            except ValueError as err:
                raise ValueError('E#{} - {}'.format(i, err))
            crosssec = self.get_element_crosssec(i)
            mat = self.get_element_material(i)
            K_loc = create_local_stiffness_matrix(L, crosssec.A, crosssec.Jx,\
                crosssec.Iy, crosssec.Iz, mat.E, mat.mu,\
                cr1=cr1, cr2=cr2)
            R_LG = turn_diagblock(R3) # 12x12
            K_G = ((R_LG.T).dot(K_loc)).dot(R_LG)
            self._element_k_list[i] = K_G
            self._rot_list[i] = R_LG

    ################################
    # private functions - solve-time vector/matrix assembly
    def _create_self_weight_load(self, exist_e_ids):
        assert len(self._element_gravity_nload_list) > 0, 'Call `set_self_weight_load` first to init the load!'
        self_weight_P = np.zeros(self.nV * 6)
        for e in exist_e_ids:
            Qe = self._element_gravity_nload_list[e]
            for dof_j in range(self._id_map.shape[1]):
                self_weight_P[self._id_map[e, dof_j]] += Qe[dof_j]
        return self_weight_P
    
    def _create_uniformly_distributed_lumped_load(self, exist_e_ids):
        assert len(self._element_lumped_nload_list) > 0, 'Call `set_load` first to init the load!'
        element_lump_P = np.zeros(self.nV * 6)
        for e in exist_e_ids:
            Qe = self._element_lumped_nload_list[e]
            for dof_j in range(self._id_map.shape[1]):
                element_lump_P[self._id_map[e, dof_j]] += Qe[dof_j]
        return element_lump_P